# SPDX-FileCopyrightText: 2019 Free Software Foundation Europe e.V. <https://fsfe.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Entry module for reuse."""

import sys

if __name__ == "__main__":
    from ._main import main

    sys.exit(main())
